@@ -14,9 +14,10 @@
 #include "iotype.h"
 
 unsigned long debug_enable_flags;
+unsigned long worker_thread_pool_size;
 
 static struct kmem_cache *iscsi_cmnd_cache;
-static char dummy_data[1024];
+static u8 dummy_data[PAGE_SIZE];
 
 static int ctr_major;
 static char ctr_name[] = "ietctl";
@@ -138,7 +139,7 @@ static void iscsi_scsi_dequeuecmnd(struct iscsi_cmnd *cmnd)
 /**
  * create a new command.
  *
- * iscsi_cmnd_create - 
+ * iscsi_cmnd_create -
  * @conn: ptr to connection (for i/o)
  *
  * @return    ptr to command or NULL
@@ -172,7 +173,7 @@ struct iscsi_cmnd *cmnd_alloc(struct iscsi_conn *conn, int req)
 /**
  * create a new command used as response.
  *
- * iscsi_cmnd_create_rsp_cmnd - 
+ * iscsi_cmnd_create_rsp_cmnd -
  * @cmnd: ptr to request command
  *
  * @return    ptr to response command or NULL
@@ -301,25 +302,6 @@ static struct iscsi_cmnd *create_scsi_rsp(struct iscsi_cmnd *req)
 	struct iscsi_cmnd *rsp;
 	struct iscsi_scsi_cmd_hdr *req_hdr = cmnd_hdr(req);
 	struct iscsi_scsi_rsp_hdr *rsp_hdr;
-
-	rsp = iscsi_cmnd_create_rsp_cmnd(req, 1);
-
-	rsp_hdr = (struct iscsi_scsi_rsp_hdr *)&rsp->pdu.bhs;
-	rsp_hdr->opcode = ISCSI_OP_SCSI_RSP;
-	rsp_hdr->flags = ISCSI_FLG_FINAL;
-	rsp_hdr->response = ISCSI_RESPONSE_COMMAND_COMPLETED;
-	rsp_hdr->cmd_status = SAM_STAT_GOOD;
-	rsp_hdr->itt = req_hdr->itt;
-
-	return rsp;
-}
-
-static struct iscsi_cmnd *create_sense_rsp(struct iscsi_cmnd *req,
-					   u8 sense_key, u8 asc, u8 ascq)
-{
-	struct iscsi_cmnd *rsp;
-	struct iscsi_scsi_rsp_hdr *rsp_hdr;
-	struct tio *tio;
 	struct iscsi_sense_data *sense;
 
 	rsp = iscsi_cmnd_create_rsp_cmnd(req, 1);
@@ -328,79 +310,93 @@ static struct iscsi_cmnd *create_sense_rsp(struct iscsi_cmnd *req,
 	rsp_hdr->opcode = ISCSI_OP_SCSI_RSP;
 	rsp_hdr->flags = ISCSI_FLG_FINAL;
 	rsp_hdr->response = ISCSI_RESPONSE_COMMAND_COMPLETED;
-	rsp_hdr->cmd_status = SAM_STAT_CHECK_CONDITION;
-	rsp_hdr->itt = cmnd_hdr(req)->itt;
+	rsp_hdr->cmd_status = req->status;
+	rsp_hdr->itt = req_hdr->itt;
 
-	tio = rsp->tio = tio_alloc(1);
-	sense = (struct iscsi_sense_data *) page_address(tio->pvec[0]);
-	assert(sense);
-	clear_page(sense);
-	sense->length = cpu_to_be16(14);
-	sense->data[0] = 0xf0;
-	sense->data[2] = sense_key;
-	sense->data[7] = 6;	// Additional sense length
-	sense->data[12] = asc;
-	sense->data[13] = ascq;
+	if (req->status == SAM_STAT_CHECK_CONDITION) {
+		assert(!rsp->tio);
+		rsp->tio = tio_alloc(1);
+		sense = (struct iscsi_sense_data *)
+			page_address(rsp->tio->pvec[0]);
 
-	rsp->pdu.datasize = sizeof(struct iscsi_sense_data) + 14;
-	tio->size = (rsp->pdu.datasize + 3) & -4;
-	tio->offset = 0;
+		assert(sense);
+		clear_page(sense);
+		sense->length = cpu_to_be16(IET_SENSE_BUF_SIZE);
+
+		memcpy(sense->data, req->sense_buf, IET_SENSE_BUF_SIZE);
+		rsp->pdu.datasize = sizeof(struct iscsi_sense_data) +
+			IET_SENSE_BUF_SIZE;
+
+		rsp->tio->size = (rsp->pdu.datasize + 3) & -4;
+		rsp->tio->offset = 0;
+	}
 
 	return rsp;
 }
 
-void send_scsi_rsp(struct iscsi_cmnd *req, int (*func)(struct iscsi_cmnd *))
+void iscsi_cmnd_set_sense(struct iscsi_cmnd *cmnd, u8 sense_key, u8 asc,
+			  u8 ascq)
+{
+	cmnd->status = SAM_STAT_CHECK_CONDITION;
+
+	cmnd->sense_buf[0] = 0xf0;
+	cmnd->sense_buf[2] = sense_key;
+	cmnd->sense_buf[7] = 6;	// Additional sense length
+	cmnd->sense_buf[12] = asc;
+	cmnd->sense_buf[13] = ascq;
+}
+
+static struct iscsi_cmnd *create_sense_rsp(struct iscsi_cmnd *req,
+					   u8 sense_key, u8 asc, u8 ascq)
+{
+	iscsi_cmnd_set_sense(req, sense_key, asc, ascq);
+	return create_scsi_rsp(req);
+}
+
+void send_scsi_rsp(struct iscsi_cmnd *req, void (*func)(struct iscsi_cmnd *))
 {
 	struct iscsi_cmnd *rsp;
 	struct iscsi_scsi_rsp_hdr *rsp_hdr;
 	u32 size;
-	int ret = func(req);
 
-	switch (ret) {
-	case 0:
-	case -EBUSY:
-		rsp = create_scsi_rsp(req);
+	func(req);
+	rsp = create_scsi_rsp(req);
+
+	switch (req->status) {
+	case SAM_STAT_GOOD:
+	case SAM_STAT_RESERVATION_CONFLICT:
 		rsp_hdr = (struct iscsi_scsi_rsp_hdr *) &rsp->pdu.bhs;
 		if ((size = cmnd_read_size(req)) != 0) {
 			rsp_hdr->flags |= ISCSI_FLG_RESIDUAL_UNDERFLOW;
 			rsp_hdr->residual_count = cpu_to_be32(size);
 		}
-		if (ret == -EBUSY)
-			rsp_hdr->cmd_status = SAM_STAT_RESERVATION_CONFLICT;
-		break;
-	case -EIO:
-		/* Medium Error/Write Fault */
-		rsp = create_sense_rsp(req, MEDIUM_ERROR, 0x03, 0x0);
 		break;
 	default:
-		rsp = create_sense_rsp(req, ILLEGAL_REQUEST, 0x24, 0x0);
+		break;
 	}
+
 	iscsi_cmnd_init_write(rsp);
 }
 
-void send_data_rsp(struct iscsi_cmnd *req, int (*func)(struct iscsi_cmnd *))
+void send_data_rsp(struct iscsi_cmnd *req, void (*func)(struct iscsi_cmnd *))
 {
 	struct iscsi_cmnd *rsp;
 
-	switch (func(req)) {
-	case 0:
+	func(req);
+
+	if (req->status == SAM_STAT_GOOD)
 		do_send_data_rsp(req);
-		return;
-	case -EIO:
-		/* Medium Error/Unrecovered Read Error */
-		rsp = create_sense_rsp(req, MEDIUM_ERROR, 0x11, 0x0);
-		break;
-	default:
-		rsp = create_sense_rsp(req, ILLEGAL_REQUEST, 0x24, 0x0);
+	else {
+		rsp = create_scsi_rsp(req);
+		iscsi_cmnd_init_write(rsp);
 	}
-	iscsi_cmnd_init_write(rsp);
 }
 
 /**
  * Free a command.
  * Also frees the additional header.
  *
- * iscsi_cmnd_remove - 
+ * iscsi_cmnd_remove -
  * @cmnd: ptr to command
  */
 
@@ -726,13 +722,18 @@ static void set_offset_and_length(struct iet_volume *lu, u8 *cmd, loff_t *off, u
 	case READ_10:
 	case WRITE_10:
 	case WRITE_VERIFY:
-		*off = be32_to_cpu(*(u32 *)&cmd[2]);
+		*off = (u32)cmd[2] << 24 | (u32)cmd[3] << 16 |
+			(u32)cmd[4] << 8 | (u32)cmd[5];
 		*len = (cmd[7] << 8) + cmd[8];
 		break;
 	case READ_16:
 	case WRITE_16:
-		*off = be64_to_cpu(*(u64 *)&cmd[2]);
-		*len = be32_to_cpu(*(u32 *)&cmd[10]);
+		*off = (u64)cmd[2] << 56 | (u64)cmd[3] << 48 |
+			(u64)cmd[4] << 40 | (u64)cmd[5] << 32 |
+			(u64)cmd[6] << 24 | (u64)cmd[7] << 16 |
+			(u64)cmd[8] << 8 | (u64)cmd[9];
+		*len = (u32)cmd[10] << 24 | (u32)cmd[11] << 16 |
+			(u32)cmd[12] << 8 | (u32)cmd[13];
 		break;
 	default:
 		BUG();
@@ -1153,7 +1154,7 @@ static int target_reset(struct iscsi_cmnd *req, u32 lun, int all)
 	struct iscsi_session *session;
 	struct iscsi_conn *conn;
 	struct iscsi_cmnd *cmnd, *tmp;
-	struct iet_volume *volumes;
+	struct iet_volume *volume;
 
 	list_for_each_entry(session, &target->session_list, list) {
 		list_for_each_entry(conn, &session->conn_list, list) {
@@ -1169,10 +1170,15 @@ static int target_reset(struct iscsi_cmnd *req, u32 lun, int all)
 		}
 	}
 
-	list_for_each_entry(volumes, &target->volumes, list)
-		if (all || volumes->lun == lun)
+	list_for_each_entry(volume, &target->volumes, list) {
+		if (all || volume->lun == lun) {
 			/* force release */
-			volume_release(volumes, 0, 1);
+			volume_release(volume, 0, 1);
+			/* power-on, reset, or bus device reset occurred */
+			ua_establish_for_all_sessions(target, volume->lun,
+						      0x29, 0x0);
+		}
+	}
 
 	return 0;
 }
@@ -1191,10 +1197,56 @@ static void task_set_abort(struct iscsi_cmnd *req)
 	}
 }
 
+static inline char *tmf_desc(int fun)
+{
+	static char *tmf_desc[] = {
+		"Unknown Function",
+		"Abort Task",
+		"Abort Task Set",
+		"Clear ACA",
+		"Clear Task Set",
+		"Logical Unit Reset",
+		"Target Warm Reset",
+		"Target Cold Reset",
+		"Task Reassign",
+        };
+
+	if ((fun < ISCSI_FUNCTION_ABORT_TASK) ||
+				(fun > ISCSI_FUNCTION_TASK_REASSIGN))
+		fun = 0;
+
+	return tmf_desc[fun];
+}
+
+static inline char *rsp_desc(int rsp)
+{
+	static char *rsp_desc[] = {
+		"Function Complete",
+		"Unknown Task",
+		"Unknown LUN",
+		"Task Allegiant",
+		"Failover Unsupported",
+		"Function Unsupported",
+		"No Authorization",
+		"Function Rejected",
+		"Unknown Response",
+	};
+
+	if (((rsp < ISCSI_RESPONSE_FUNCTION_COMPLETE) ||
+			(rsp > ISCSI_RESPONSE_NO_AUTHORIZATION)) &&
+			(rsp != ISCSI_RESPONSE_FUNCTION_REJECTED))
+		rsp = 8;
+	else if (rsp == ISCSI_RESPONSE_FUNCTION_REJECTED)
+		rsp = 7;
+
+	return rsp_desc[rsp];
+}
+
 static void execute_task_management(struct iscsi_cmnd *req)
 {
 	struct iscsi_conn *conn = req->conn;
-	struct iscsi_target *target = conn->session->target;
+	struct iscsi_session *session = conn->session;
+	struct iscsi_target *target = session->target;
 	struct iscsi_cmnd *rsp;
 	struct iscsi_task_mgt_hdr *req_hdr = (struct iscsi_task_mgt_hdr *)&req->pdu.bhs;
 	struct iscsi_task_rsp_hdr *rsp_hdr;
@@ -1209,8 +1261,6 @@ static void execute_task_management(struct iscsi_cmnd *req)
 	rsp_hdr->itt = req_hdr->itt;
 	rsp_hdr->response = ISCSI_RESPONSE_FUNCTION_COMPLETE;
 
-	eprintk("%x %d %x\n", cmnd_itt(req), function, req_hdr->rtt);
-
 	switch (function) {
 	case ISCSI_FUNCTION_ABORT_TASK:
 	case ISCSI_FUNCTION_ABORT_TASK_SET:
@@ -1255,6 +1305,11 @@ static void execute_task_management(struct iscsi_cmnd *req)
 		break;
 	}
 out:
+	iprintk("%s (%02x) issued on tid:%d lun:%d by sid:%llu (%s)\n",
+				tmf_desc(function), function, target->tid,
+				translate_lun(req_hdr->lun), session->sid,
+				rsp_desc(rsp_hdr->response));
+
 	iscsi_cmnd_init_write(rsp);
 }
 
@@ -1520,7 +1575,7 @@ void cmnd_tx_end(struct iscsi_cmnd *cmnd)
  * This functions reorders the commands.
  * Called from the read thread.
  *
- * iscsi_session_push_cmnd - 
+ * iscsi_session_push_cmnd -
  * @cmnd: ptr to command
  */
 
@@ -1677,6 +1732,8 @@ void cmnd_rx_end(struct iscsi_cmnd *cmnd)
 
 static void iscsi_exit(void)
 {
+	wthread_module_exit();
+
 	unregister_chrdev(ctr_major, ctr_name);
 
 	iet_procfs_exit();
@@ -1687,6 +1744,8 @@ static void iscsi_exit(void)
 
 	iotype_exit();
 
+	ua_exit();
+
 	if (iscsi_cmnd_cache)
 		kmem_cache_destroy(iscsi_cmnd_cache);
 }
@@ -1708,17 +1767,23 @@ static int iscsi_init(void)
 	if ((err = event_init()) < 0)
 		goto err;
 
-	iscsi_cmnd_cache = kmem_cache_create("iscsi_cmnd", sizeof(struct iscsi_cmnd),
-					     0, 0, NULL, NULL);
+	iscsi_cmnd_cache = KMEM_CACHE(iscsi_cmnd, 0);
 	if (!iscsi_cmnd_cache)
 		goto err;
 
+	err = ua_init();
+	if (err < 0)
+		goto err;
+
 	if ((err = tio_init()) < 0)
 		goto err;
 
 	if ((err = iotype_init()) < 0)
 		goto err;
 
+	if ((err = wthread_module_init()) < 0)
+		goto err;
+
 	return 0;
 
 err:
@@ -1726,9 +1791,19 @@ err:
 	return err;
 }
 
-module_param(debug_enable_flags, ulong, S_IRUGO);
+module_param(worker_thread_pool_size, ulong, S_IRUGO);
+MODULE_PARM_DESC(worker_thread_pool_size,
+		 "Size of the worker thread pool "
+		 "(0 = dedicated threads per target (default))");
+
+module_param(debug_enable_flags, ulong, S_IRUGO | S_IWUSR);
+MODULE_PARM_DESC(debug_enable_flags,
+		 "debug bitmask, low bits (0 ... 8) used, see iscsi_dbg.h");
 
 module_init(iscsi_init);
 module_exit(iscsi_exit);
 
+MODULE_VERSION(IET_VERSION_STRING);
 MODULE_LICENSE("GPL");
+MODULE_DESCRIPTION("iSCSI Enterprise Target");
+MODULE_AUTHOR("IET development team <iscsitarget-devel@lists.sourceforge.net>");
