@@ -2,8 +2,7 @@
 #define CONFIG_H
 
 struct config_operations {
-	int (*init) (char *, char **, int *);
-	int (*default_load) (char *);
+	void (*init) (char *, char **, int *);
 	int (*target_add) (u32 *, char *);
 	int (*target_stop) (u32);
 	int (*target_del) (u32);
@@ -15,7 +14,8 @@ struct config_operations {
 	int (*account_del) (u32, int, char *);
 	int (*account_query) (u32, int, char *, char *);
 	int (*account_list) (u32, int, u32 *, u32 *, char *, size_t);
-	int (*initiator_access) (u32, int);
+	int (*initiator_allow) (u32, int, char *);
+	int (*target_allow) (u32, struct sockaddr *);
 };
 
 extern struct config_operations *cops;
