@@ -13,6 +13,7 @@
 #include <unistd.h>
 #include <getopt.h>
 #include <netdb.h>
+#include <signal.h>
 
 #include <sys/poll.h>
 #include <sys/socket.h>
@@ -28,24 +29,10 @@
 #include "iscsid.h"
 #include "ietadm.h"
 
-#define LISTEN_MAX		8
-#define INCOMING_MAX		32
-
-enum {
-	POLL_LISTEN,
-	POLL_IPC = POLL_LISTEN + LISTEN_MAX,
-	POLL_NL,
-	POLL_ISNS,
-	POLL_SCN_LISTEN,
-	POLL_SCN,
-	POLL_INCOMING,
-	POLL_MAX = POLL_INCOMING + INCOMING_MAX,
-};
-
 static char* server_address;
 uint16_t server_port = ISCSI_LISTEN_PORT;
 
-static struct pollfd poll_array[POLL_MAX];
+struct pollfd poll_array[POLL_MAX];
 static struct connection *incoming[INCOMING_MAX];
 static int incoming_cnt;
 int ctrl_fd, ipc_fd, nl_fd;
@@ -66,7 +53,7 @@ static struct option const long_options[] =
 	{0, 0, 0, 0},
 };
 
-/* This will be comfigurable by command line options */
+/* This will be configurable by command line options */
 extern struct config_operations plain_ops;
 struct config_operations *cops = &plain_ops;
 
@@ -118,7 +105,9 @@ static void create_listen_socket(struct pollfd *array)
 	hints.ai_flags = AI_PASSIVE;
 
 	if (getaddrinfo(server_address, servname, &hints, &res0)) {
-		log_error("unable to get address info (%s)!", strerror(errno));
+		log_error("unable to get address info (%s)!",
+			(errno == EAI_SYSTEM) ? strerror(errno) :
+						gai_strerror(errno));
 		exit(1);
 	}
 
@@ -135,10 +124,12 @@ static void create_listen_socket(struct pollfd *array)
 		if (setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &opt, sizeof(opt)))
 			log_warning("unable to set SO_KEEPALIVE on server socket (%s)!",
 				    strerror(errno));
+
 		opt = 1;
 		if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)))
 			log_warning("unable to set SO_REUSEADDR on server socket (%s)!",
 				    strerror(errno));
+
 		opt = 1;
 		if (res->ai_family == AF_INET6 &&
 		    setsockopt(sock, IPPROTO_IPV6, IPV6_V6ONLY, &opt, sizeof(opt)))
@@ -413,7 +404,7 @@ void event_loop(int timeout)
 
 		conn_close:
 			if (conn->state == STATE_CLOSE) {
-				log_debug(0, "connection closed");
+				log_debug(1, "connection closed");
 				conn_free_pdu(conn);
 				conn_free(conn);
 				close(pollfd->fd);
@@ -434,6 +425,10 @@ int main(int argc, char **argv)
 	char *isns = NULL;
 	int isns_ac = 0;
 
+	/* otherwise we would die in some later write() during the event_loop
+	 * instead of getting EPIPE! */
+	signal(SIGPIPE, SIG_IGN);
+
 	while ((ch = getopt_long(argc, argv, "c:fd:s:u:g:a:p:vh", long_options, &longindex)) >= 0) {
 		switch (ch) {
 		case 'c':
@@ -534,8 +529,6 @@ int main(int argc, char **argv)
 	if (isns)
 		timeout = isns_init(isns, isns_ac);
 
-	cops->default_load(config);
-
 	if (gid && setgid(gid) < 0)
 		perror("setgid");
 
