@@ -16,6 +16,14 @@
 
 #include "iscsid.h"
 
+static u32 ttt;
+
+static u32 get_next_ttt(struct connection *conn __attribute__((unused)))
+{
+	ttt += 1;
+	return (ttt == ISCSI_RESERVED_TAG) ? ++ttt : ttt;
+}
+
 static struct iscsi_key login_keys[] = {
 	{"InitiatorName",},
 	{"InitiatorAlias",},
@@ -81,31 +89,91 @@ static char *next_key(char **data, int *datasize, char **value)
 	return key;
 }
 
+static struct buf_segment * conn_alloc_buf_segment(struct connection *conn,
+						   size_t sz)
+{
+	struct buf_segment *seg = malloc(sizeof *seg + sz);
+
+	if (seg) {
+		seg->len = 0;
+		memset(seg->data, 0x0, sz);
+		list_add_tail(&seg->entry, &conn->rsp_buf_list);
+		log_debug(2, "alloc'ed new buf_segment");
+	}
+
+	return seg;
+}
+
+
 void text_key_add(struct connection *conn, char *key, char *value)
 {
+	struct buf_segment *seg;
 	int keylen = strlen(key);
 	int valuelen = strlen(value);
 	int len = keylen + valuelen + 2;
-	char *buffer;
+	int off = 0;
+	int sz = 0;
+	int stage = 0;
+	size_t data_sz;
 
-	if (!conn->rsp.datasize) {
-		if (!conn->rsp_buffer)
-			conn->rsp_buffer = malloc(INCOMING_BUFSIZE);
-		conn->rsp.data = conn->rsp_buffer;
+	data_sz = (conn->state == STATE_FULL) ?
+		conn->session_param[key_max_xmit_data_length].val :
+		INCOMING_BUFSIZE;
+
+	seg = list_empty(&conn->rsp_buf_list) ? NULL :
+		list_entry(conn->rsp_buf_list.q_back, struct buf_segment,
+			   entry);
+
+	while (len) {
+		if (!seg || seg->len == data_sz) {
+			seg = conn_alloc_buf_segment(conn, data_sz);
+			if (!seg) {
+				log_error("Failed to alloc text buf segment\n");
+				conn_free_rsp_buf_list(conn);
+				break;
+			}
+		}
+		switch (stage) {
+		case 0:
+			sz = min_t(int, data_sz - seg->len, keylen - off);
+			strncpy(seg->data + seg->len, key + off, sz);
+			if (sz == data_sz - seg->len) {
+				off += sz;
+				if (keylen - off == 0) {
+					off = 0;
+					stage++;
+				}
+			} else {
+				off = 0;
+				stage++;
+			}
+			break;
+		case 1:
+			seg->data[seg->len] = '=';
+			off = 0;
+			sz = 1;
+			stage++;
+			break;
+		case 2:
+			sz = min_t(int, data_sz - seg->len, valuelen - off);
+			strncpy(seg->data + seg->len, value + off, sz);
+			off += sz;
+			if (valuelen - off == 0) {
+				off = 0;
+				stage++;
+			}
+			break;
+		case 3:
+			seg->data[seg->len] = 0;
+			sz = 1;
+			break;
+		}
+
+		log_debug(1, "wrote: %s", seg->data + seg->len);
+
+		seg->len += sz;
+		len -= sz;
 	}
-	if (conn->rwsize + len > INCOMING_BUFSIZE) {
-		log_warning("Dropping key (%s=%s)", key, value);
-		return;
-	}
-
-	buffer = conn->rsp_buffer;
-	buffer += conn->rsp.datasize;
-	conn->rsp.datasize += len;
-
-	strcpy(buffer, key);
-	buffer += keylen;
-	*buffer++ = '=';
-	strcpy(buffer, value);
 }
 
 static void text_key_add_reject(struct connection *conn, char *key)
@@ -323,6 +391,10 @@ static void login_start(struct connection *conn)
 	struct iscsi_login_req_hdr *req = (struct iscsi_login_req_hdr *)&conn->req.bhs;
 	struct iscsi_login_rsp_hdr *rsp = (struct iscsi_login_rsp_hdr *)&conn->rsp.bhs;
 	char *name, *alias, *session_type, *target_name;
+	struct sockaddr_storage ss;
+	socklen_t slen = sizeof(struct sockaddr_storage);
+
+	memset(&ss, 0, sizeof(ss));
 
 	conn->cid = be16_to_cpu(req->cid);
 	conn->sid.id64 = req->sid.id64;
@@ -348,6 +420,8 @@ static void login_start(struct connection *conn)
 	conn->auth_method = -1;
 	conn->session_type = SESSION_NORMAL;
 
+	getsockname(conn->fd, (struct sockaddr *) &ss, &slen);
+
 	if (session_type) {
 		if (!strcmp(session_type, "Discovery"))
 			conn->session_type = SESSION_DISCOVERY;
@@ -367,9 +441,10 @@ static void login_start(struct connection *conn)
 			return;
 		}
 
-		if (!(conn->tid = target_find_by_name(target_name)) ||
-		    cops->initiator_access(conn->tid, conn->fd) ||
-		    isns_scn_access(conn->tid, conn->fd, name)) {
+		if (!(conn->tid = target_find_by_name(target_name))
+			|| !cops->initiator_allow(conn->tid, conn->fd, name)
+			|| !cops->target_allow(conn->tid, (struct sockaddr *) &ss)
+			|| !isns_scn_allow(conn->tid, name)) {
 			rsp->status_class = ISCSI_STATUS_INITIATOR_ERR;
 			rsp->status_detail = ISCSI_STATUS_TGT_NOT_FOUND;
 			conn->state = STATE_EXIT;
@@ -385,7 +460,12 @@ static void login_start(struct connection *conn)
 /* 			return; */
 /* 		} */
 
-		ki->param_get(conn->tid, 0, key_session, conn->session_param);
+		if (ki->param_get(conn->tid, 0, key_session,
+				  conn->session_param)) {
+			rsp->status_class = ISCSI_STATUS_TARGET_ERR;
+			rsp->status_detail = ISCSI_STATUS_SVC_UNAVAILABLE;
+			conn->state = STATE_EXIT;
+		}
 	}
 	conn->exp_cmd_sn = be32_to_cpu(req->cmd_sn);
 	log_debug(1, "exp_cmd_sn: %d,%d", conn->exp_cmd_sn, req->cmd_sn);
@@ -407,6 +487,36 @@ static void login_finish(struct connection *conn)
 	}
 }
 
+static void cmnd_reject(struct connection *conn, u8 reason)
+{
+	struct iscsi_reject_hdr *rej =
+		(struct iscsi_reject_hdr *)&conn->rsp.bhs;
+	size_t data_sz = sizeof(struct iscsi_hdr);
+	struct buf_segment *seg;
+
+	conn_free_rsp_buf_list(conn);
+	seg = conn_alloc_buf_segment(conn, data_sz);
+
+	memset(rej, 0x0, sizeof *rej);
+	rej->opcode = ISCSI_OP_REJECT_MSG;
+	rej->reason = reason;
+	rej->ffffffff = ISCSI_RESERVED_TAG;
+	rej->flags |= ISCSI_FLG_FINAL;
+
+	rej->stat_sn = cpu_to_be32(conn->stat_sn++);
+	rej->exp_cmd_sn = cpu_to_be32(conn->exp_cmd_sn);
+	conn->max_cmd_sn = conn->exp_cmd_sn + 1;
+	rej->max_cmd_sn = cpu_to_be32(conn->max_cmd_sn);
+
+	if (!seg) {
+		log_error("Failed to alloc data segment for Reject PDU\n");
+		return;
+	}
+
+	memcpy(seg->data, &conn->req.bhs, data_sz);
+	seg->len = data_sz;
+}
+
 static int cmnd_exec_auth(struct connection *conn)
 {
        int res;
@@ -435,7 +545,8 @@ static void cmnd_exec_login(struct connection *conn)
 	memset(rsp, 0, BHS_SIZE);
 	if ((req->opcode & ISCSI_OPCODE_MASK) != ISCSI_OP_LOGIN_CMD ||
 	    !(req->opcode & ISCSI_OP_IMMEDIATE)) {
-		//reject
+		cmnd_reject(conn, ISCSI_REASON_PROTOCOL_ERROR);
+		return;
 	}
 
 	rsp->opcode = ISCSI_OP_LOGIN_RSP;
@@ -569,6 +680,17 @@ static void cmnd_exec_login(struct connection *conn)
 		rsp->flags |= nsg | (stay ? 0 : ISCSI_FLG_TRANSIT);
 	}
 
+	/*
+	 * TODO: support Logical Text Data Segments > INCOMING_BUFSIZE (i.e.
+	 * key=value pairs spanning several PDUs) during login phase
+	 */
+	if (!list_empty(&conn->rsp_buf_list) &&
+	    !list_length_is_one(&conn->rsp_buf_list)) {
+		log_error("Target error: \'key=value\' pairs spanning several "
+			  "Login PDUs are not implemented, yet\n");
+		goto target_err;
+	}
+
 	rsp->sid = conn->sid;
 	rsp->stat_sn = cpu_to_be32(conn->stat_sn++);
 	rsp->exp_cmd_sn = cpu_to_be32(conn->exp_cmd_sn);
@@ -587,6 +709,12 @@ auth_err:
 	rsp->status_detail = ISCSI_STATUS_AUTH_FAILED;
 	conn->state = STATE_EXIT;
 	return;
+ target_err:
+	rsp->flags = 0;
+	rsp->status_class = ISCSI_STATUS_TARGET_ERR;
+	rsp->status_detail = ISCSI_STATUS_TARGET_ERROR;
+	conn->state = STATE_EXIT;
+	return;
 }
 
 static void text_scan_text(struct connection *conn)
@@ -599,34 +727,10 @@ static void text_scan_text(struct connection *conn)
 
 	while ((key = next_key(&data, &datasize, &value))) {
 		if (!strcmp(key, "SendTargets")) {
-			struct sockaddr_storage ss;
-			socklen_t slen, blen;
-			char *p, buf[NI_MAXHOST + 128];
-
 			if (value[0] == 0)
 				continue;
 
-			p = buf;
-			blen = sizeof(buf);
-
-			slen = sizeof(ss);
-			getsockname(conn->fd, (struct sockaddr *) &ss, &slen);
-			if (ss.ss_family == AF_INET6) {
-				*p++ = '[';
-				blen--;
-			}
-
-			slen = sizeof(ss);
-			getnameinfo((struct sockaddr *) &ss, slen, p, blen,
-				    NULL, 0, NI_NUMERICHOST);
-
-			p = buf + strlen(buf);
-
-			if (ss.ss_family == AF_INET6)
-				 *p++ = ']';
-
-			sprintf(p, ":%d,1", server_port);
-			target_list_build(conn, buf,
+			target_list_build(conn,
 					  strcmp(value, "All") ? value : NULL);
 		} else
 			text_key_add(conn, key, "NotUnderstood");
@@ -640,23 +744,37 @@ static void cmnd_exec_text(struct connection *conn)
 
 	memset(rsp, 0, BHS_SIZE);
 
-	if (be32_to_cpu(req->ttt) != 0xffffffff) {
-		/* reject */;
-	}
 	rsp->opcode = ISCSI_OP_TEXT_RSP;
 	rsp->itt = req->itt;
-	//rsp->ttt = rsp->ttt;
-	rsp->ttt = 0xffffffff;
 	conn->exp_cmd_sn = be32_to_cpu(req->cmd_sn);
 	if (!(req->opcode & ISCSI_OP_IMMEDIATE))
 		conn->exp_cmd_sn++;
 
 	log_debug(1, "Text request: %d", conn->state);
-	text_scan_text(conn);
 
-	if (req->flags & ISCSI_FLG_FINAL)
+	if (req->ttt == ISCSI_RESERVED_TAG) {
+		conn_free_rsp_buf_list(conn);
+		text_scan_text(conn);
+		if (!list_empty(&conn->rsp_buf_list) &&
+		    !list_length_is_one(&conn->rsp_buf_list))
+			conn->ttt = get_next_ttt(conn);
+		else
+			conn->ttt = ISCSI_RESERVED_TAG;
+	} else if (list_empty(&conn->rsp_buf_list) || conn->ttt != req->ttt) {
+		log_error("Rejecting unexpected text request. TTT recv %#x, "
+			  "expected %#x; %stext segments queued\n",
+			  req->ttt, conn->ttt, list_empty(&conn->rsp_buf_list) ?
+			  "no " : "");
+		cmnd_reject(conn, ISCSI_REASON_INVALID_PDU_FIELD);
+		return;
+	}
+
+	if (list_empty(&conn->rsp_buf_list) ||
+	    list_length_is_one(&conn->rsp_buf_list))
 		rsp->flags = ISCSI_FLG_FINAL;
 
+	rsp->ttt = conn->ttt;
+
 	rsp->stat_sn = cpu_to_be32(conn->stat_sn++);
 	rsp->exp_cmd_sn = cpu_to_be32(conn->exp_cmd_sn);
 	conn->max_cmd_sn = conn->exp_cmd_sn + 1;
@@ -684,47 +802,68 @@ static void cmnd_exec_logout(struct connection *conn)
 
 int cmnd_execute(struct connection *conn)
 {
-	int res = 1;
+	struct buf_segment *seg;
+	struct iscsi_login_rsp_hdr *login_rsp;
 
 	switch (conn->req.bhs.opcode & ISCSI_OPCODE_MASK) {
 	case ISCSI_OP_LOGIN_CMD:
-		//if conn->state == STATE_FULL -> reject
+		if (conn->state == STATE_FULL) {
+			cmnd_reject(conn, ISCSI_REASON_PROTOCOL_ERROR);
+			break;
+		}
 		cmnd_exec_login(conn);
-		conn->rsp.bhs.ahslength = conn->rsp.ahssize / 4;
-		conn->rsp.bhs.datalength[0] = conn->rsp.datasize >> 16;
-		conn->rsp.bhs.datalength[1] = conn->rsp.datasize >> 8;
-		conn->rsp.bhs.datalength[2] = conn->rsp.datasize;
-		log_pdu(2, &conn->rsp);
+		login_rsp = (struct iscsi_login_rsp_hdr *) &conn->rsp.bhs;
+		if (login_rsp->status_class)
+			conn_free_rsp_buf_list(conn);
 		break;
 	case ISCSI_OP_TEXT_CMD:
-		//if conn->state != STATE_FULL -> reject
-		cmnd_exec_text(conn);
-		conn->rsp.bhs.ahslength = conn->rsp.ahssize / 4;
-		conn->rsp.bhs.datalength[0] = conn->rsp.datasize >> 16;
-		conn->rsp.bhs.datalength[1] = conn->rsp.datasize >> 8;
-		conn->rsp.bhs.datalength[2] = conn->rsp.datasize;
-		log_pdu(2, &conn->rsp);
+		if (conn->state != STATE_FULL)
+			cmnd_reject(conn, ISCSI_REASON_PROTOCOL_ERROR);
+		else
+			cmnd_exec_text(conn);
 		break;
 	case ISCSI_OP_LOGOUT_CMD:
-		//if conn->state != STATE_FULL -> reject
-		cmnd_exec_logout(conn);
-		conn->rsp.bhs.ahslength = conn->rsp.ahssize / 4;
-		conn->rsp.bhs.datalength[0] = conn->rsp.datasize >> 16;
-		conn->rsp.bhs.datalength[1] = conn->rsp.datasize >> 8;
-		conn->rsp.bhs.datalength[2] = conn->rsp.datasize;
-		log_pdu(2, &conn->rsp);
+		if (conn->state != STATE_FULL)
+			cmnd_reject(conn, ISCSI_REASON_PROTOCOL_ERROR);
+		else
+			cmnd_exec_logout(conn);
 		break;
 	default:
-		//reject
-		res = 0;
-		break;
+		cmnd_reject(conn, ISCSI_REASON_UNSUPPORTED_COMMAND);
+		return 0;
 	}
 
-	return res;
+	if (!list_empty(&conn->rsp_buf_list)) {
+		seg = list_entry(conn->rsp_buf_list.q_forw,
+				 struct buf_segment, entry);
+		list_del_init(&seg->entry);
+		conn->rsp.datasize = seg->len;
+		conn->rsp.data = seg->data;
+	} else {
+		conn->rsp.datasize = 0;
+		conn->rsp.data = NULL;
+	}
+
+	conn->rsp.bhs.ahslength = conn->rsp.ahssize / 4;
+	conn->rsp.bhs.datalength[0] = conn->rsp.datasize >> 16;
+	conn->rsp.bhs.datalength[1] = conn->rsp.datasize >> 8;
+	conn->rsp.bhs.datalength[2] = conn->rsp.datasize;
+	log_pdu(2, &conn->rsp);
+
+	return 1;
 }
 
 void cmnd_finish(struct connection *conn)
 {
+	struct buf_segment *seg;
+
+	if (conn->rsp.data) {
+		seg = container_of(conn->rsp.data, struct buf_segment, data);
+		list_del(&seg->entry);
+		free(seg);
+		conn->rsp.data = NULL;
+	}
+
 	switch (conn->state) {
 	case STATE_EXIT:
 		conn->state = STATE_CLOSE;
