@@ -8,7 +8,9 @@
 #include <ctype.h>
 #include <dirent.h>
 #include <errno.h>
+#include <limits.h>
 #include <netdb.h>
+#include <regex.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -23,8 +25,11 @@
 #include "iscsid.h"
 
 #define BUFSIZE		4096
-#define CONFIG_FILE	"/etc/ietd.conf"
-#define ACCT_CONFIG_FILE	CONFIG_FILE
+#define CONFIG_FILE	"ietd.conf"
+#define CONFIG_DIR	"/etc/iet/"
+#define INI_ALLOW_FILE	"initiators.allow"
+#define INI_DENY_FILE	"initiators.deny"
+#define TGT_ALLOW_FILE	"targets.allow"
 
 /*
  * Account configuration code
@@ -85,43 +90,6 @@ static struct __qelem *account_list_get(u32 tid, int dir)
 	return list;
 }
 
-static int plain_account_init(char *filename)
-{
-	FILE *fp;
-	char buf[BUFSIZE], *p, *q;
-	u32 tid;
-	int idx;
-
-	if (!(fp = fopen(filename, "r")))
-		return -EIO;
-
-	tid = 0;
-	while (fgets(buf, sizeof(buf), fp)) {
-		q = buf;
-		p = target_sep_string(&q);
-		if (!p || *p == '#')
-			continue;
-
-		if (!strcasecmp(p, "Target")) {
-			tid = 0;
-			if (!(p = target_sep_string(&q)))
-				continue;
-			tid = target_find_by_name(p);
-		} else if (!((idx = param_index_by_name(p, user_keys)) < 0)) {
-			char *name, *pass;
-			name = target_sep_string(&q);
-			pass = target_sep_string(&q);
-
-			if (cops->account_add(tid, idx, name, pass) < 0)
-				fprintf(stderr, "%s %s\n", name, pass);
-		}
-	}
-
-	fclose(fp);
-
-	return 0;
-}
-
 /* Return the first account if the length of name is zero */
 static struct user *account_lookup_by_name(u32 tid, int dir, char *name)
 {
@@ -168,6 +136,8 @@ static int plain_account_list(u32 tid, int dir, u32 *cnt, u32 *overflow,
 		return -ENOENT;
 
 	list_for_each_entry(user, list, ulist) {
+		if (user->tid != tid)
+			continue;
 		if (buf_sz >= ISCSI_NAME_LEN) {
 			strncpy(buf, user->name, ISCSI_NAME_LEN);
 			buf_sz -= ISCSI_NAME_LEN;
@@ -237,15 +207,16 @@ static int plain_account_add(u32 tid, int dir, char *name, char *pass)
 
 	user->tid = tid;
 	list = account_list_get(tid, dir);
-
-	if (dir == AUTH_DIR_OUTGOING && !list_empty(list)) {
-		struct user *old;
-		log_warning("Only one outgoing %s account is supported."
-			    " Replacing the old one.\n",
-			    tid ? "target" : "discovery");
-
-		old = list_entry(list->q_forw, struct user, ulist);
-		account_destroy(old);
+	if (dir == AUTH_DIR_OUTGOING) {
+		struct user *old, *tmp;
+		list_for_each_entry_safe(old, tmp, list, ulist) {
+			if (tid != old->tid)
+				continue;
+			log_warning("Only one outgoing %s account is supported."
+				    " Replacing the old one.\n",
+				    tid ? "target" : "discovery");
+			account_destroy(old);
+		}
 	}
 
 	insque(user, list);
@@ -338,25 +309,49 @@ static int address_match(struct sockaddr *sa1, struct sockaddr *sa2)
 	return 0;
 }
 
-static int __initiator_match(int fd, char *str)
+static int initiator_match(char *initiator, char *p)
 {
-	struct sockaddr_storage from;
-	struct addrinfo hints, *res;
-	socklen_t len;
-	char *p, *q;
-	int err = 0;
+	int match = 0;
+	char pattern[strlen(p) + 3];
+	regex_t re;
 
-	len = sizeof(from);
-	if (getpeername(fd, (struct sockaddr *) &from, &len) < 0)
+	snprintf(pattern, sizeof(pattern), "^%s$", p);
+
+	if (regcomp(&re, pattern, REG_NOSUB))
 		return 0;
 
+	match = !regexec(&re, initiator, (size_t) 0, NULL, 0);
+
+	regfree(&re);
+
+	return match;
+}
+
+static int __match(struct sockaddr *sa, char *initiator, char *str)
+{
+	struct addrinfo hints, *res;
+	char *p, *q;
+	int match = 0;
+
 	while ((p = strsep(&str, ","))) {
 		while (isspace(*p))
 			p++;
 
+		if (!*p)
+			continue;
+
+		q = p + (strlen(p) - 1);
+
+		while (isspace(*q))
+			*(q--) = '\0';
+
 		if (!strcmp(p, "ALL"))
 			return 1;
 
+		if (initiator)
+			if (initiator_match(initiator, p))
+				return 1;
+
 		if (*p == '[') {
 			p++;
 			if (!(q = strchr(p, ']')))
@@ -372,55 +367,64 @@ static int __initiator_match(int fd, char *str)
 		hints.ai_socktype = SOCK_STREAM;
 		hints.ai_flags = AI_NUMERICHOST;
 
-		if (getaddrinfo(p, NULL, &hints, &res) < 0)
-			return 0;
+		if (getaddrinfo(p, NULL, &hints, &res) == 0) {
+			if (q)
+				match = netmask_match(res->ai_addr, sa, q);
+			else
+				match = address_match(res->ai_addr, sa);
 
-		if (q)
-			err = netmask_match(res->ai_addr,
-					    (struct sockaddr *) &from, q);
-		else
-			err = address_match(res->ai_addr,
-					    (struct sockaddr *) &from);
+			freeaddrinfo(res);
+		}
 
-		freeaddrinfo(res);
-
-		if (err)
+		if (match)
 			break;
 	}
 
-	return err;
+	return match;
 }
 
-static int initiator_match(u32 tid, int fd, char *filename)
+static int match(u32 tid, struct sockaddr *sa, char *initiator, char *filename)
 {
-	int err = 0;
+	int err = -EPERM;
 	FILE *fp;
-	char buf[BUFSIZE], *p;
+	char buf[BUFSIZE], fname[PATH_MAX], *p, *q;
 
-	if (!(fp = fopen(filename, "r")))
-		return err;
+	snprintf(fname, sizeof(fname), "%s%s", CONFIG_DIR, filename);
+	if (!(fp = fopen(fname, "r"))) {
+		if (errno != ENOENT)
+			return -errno;
+
+		snprintf(fname, sizeof(fname), "%s%s", "/etc/", filename);
+		fp = fopen(fname, "r");
+		if (!fp)
+			return -errno;
+	}
 
 	/*
 	 * Every time we are called, we read the file. So we don't need to
 	 * implement 'reload feature'. It's slow, however, it doesn't matter.
 	 */
 	while ((p = fgets(buf, sizeof(buf), fp))) {
+		q = &buf[strlen(buf) - 1];
+
+		if (*q != '\n')
+			continue;
+		else
+			*q = '\0';
+
+		q = p;
+
+		p = target_sep_string(&q);
+
 		if (!p || *p == '#')
 			continue;
 
-		p = &buf[strlen(buf) - 1];
-		if (*p != '\n')
-			continue;
-		*p = '\0';
-
-		if (!(p = strchr(buf, ' ')))
-			continue;
-		*(p++) = '\0';
-
-		if (target_find_by_name(buf) != tid && strcmp(buf, "ALL"))
+		if (target_find_by_name(p) != tid && strcmp(p, "ALL"))
 			continue;
 
-		err = __initiator_match(fd, p);
+		if (__match(sa, initiator, q))
+			err = 0;
+
 		break;
 	}
 
@@ -428,13 +432,38 @@ static int initiator_match(u32 tid, int fd, char *filename)
 	return err;
 }
 
-static int plain_initiator_access(u32 tid, int fd)
+static int plain_initiator_allow(u32 tid, int fd, char *initiator)
 {
-	if (initiator_match(tid, fd, "/etc/initiators.deny") &&
-	    !initiator_match(tid, fd, "/etc/initiators.allow"))
-		return -EPERM;
-	else
-		return 0;
+	struct sockaddr_storage ss;
+	socklen_t slen = sizeof(struct sockaddr_storage);
+	int allow, deny;
+
+	memset(&ss, 0, sizeof(ss));
+
+	getpeername(fd, (struct sockaddr *) &ss, &slen);
+
+	allow = match(tid, (struct sockaddr *) &ss, initiator,
+						INI_ALLOW_FILE);
+	deny = match(tid, (struct sockaddr *) &ss, initiator,
+						INI_DENY_FILE);
+
+	if (deny != -ENOENT) {
+		if (!deny && allow)
+			return 0;
+		else
+			return 1;
+	}
+
+	return (allow == -ENOENT) ? 1 : !allow;
+}
+
+static int plain_target_allow(u32 tid, struct sockaddr *sa)
+{
+	int allow;
+
+	allow = match(tid, sa, NULL, TGT_ALLOW_FILE);
+
+	return (allow == -ENOENT) ? 1 : !allow;
 }
 
 /*
@@ -528,24 +557,46 @@ static int iscsi_param_partial_set(u32 tid, u64 sid, int type, int key, u32 val)
 	else
 		param = target_param;
 
+	memset(param, 0x0, (type == key_session) ?
+	       sizeof(session_param) :
+	       sizeof(target_param));
+
 	param[key].val = val;
 
 	return __plain_param_set(tid, sid, type, 1 << key, param, 0);
 }
 
-static int plain_main_init(char *filename)
+static void plain_portal_init(FILE *fp, char **isns, int *isns_ac)
+{
+	char buf[BUFSIZE];
+	char *p, *q;
+
+	while (fgets(buf, BUFSIZE, fp)) {
+		q = buf;
+		p = target_sep_string(&q);
+		if (!p || *p == '#')
+			continue;
+		if (!strcasecmp(p, "iSNSServer")) {
+			*isns = strdup(target_sep_string(&q));
+		} else if (!strcasecmp(p, "iSNSAccessControl")) {
+			char *str = target_sep_string(&q);
+			if (!strcasecmp(str, "Yes"))
+				*isns_ac = 1;
+		}
+	}
+
+	return;
+}
+
+static void plain_target_init(FILE *fp)
 {
-	FILE *config;
 	char buf[BUFSIZE];
 	char *p, *q;
 	int idx;
 	u32 tid, val;
 
-	if (!(config = fopen(filename, "r")))
-		return -errno;
-
 	tid = 0;
-	while (fgets(buf, BUFSIZE, config)) {
+	while (fgets(buf, BUFSIZE, fp)) {
 		q = buf;
 		p = target_sep_string(&q);
 		if (!p || *p == '#')
@@ -578,61 +629,87 @@ static int plain_main_init(char *filename)
 		}
 	}
 
-	fclose(config);
-	return 0;
+	return;
 }
 
-static int plain_default_load(char *params)
+static void plain_account_init(FILE *fp)
 {
-	int i, err;
+	char buf[BUFSIZE], *p, *q;
+	u32 tid;
+	int idx;
+
+	tid = 0;
+	while (fgets(buf, sizeof(buf), fp)) {
+		q = buf;
+		p = target_sep_string(&q);
+		if (!p || *p == '#')
+			continue;
+
+		if (!strcasecmp(p, "Target")) {
+			tid = 0;
+			if (!(p = target_sep_string(&q)))
+				continue;
+			tid = target_find_by_name(p);
+		} else if (!((idx = param_index_by_name(p, user_keys)) < 0)) {
+			char *name, *pass;
+			name = target_sep_string(&q);
+			pass = target_sep_string(&q);
+
+			if (plain_account_add(tid, idx, name, pass) < 0)
+				fprintf(stderr, "%s %s\n", name, pass);
+		}
+	}
+
+	return;
+}
+
+static void plain_init(char *params, char **isns, int *isns_ac)
+{
+	FILE *fp;
+	struct stat st;
+	char file1[PATH_MAX], file2[PATH_MAX];
+	int i;
 
 	for (i = 0; i < 1 << HASH_ORDER; i++) {
 		INIT_LIST_HEAD(&trgt_acct_in[i]);
 		INIT_LIST_HEAD(&trgt_acct_out[i]);
 	}
 
-	/* First, we must finish the main configuration. */
-	if ((err = plain_main_init(params ? params : CONFIG_FILE)))
-		return err;
+	snprintf(file1, sizeof(file1), "%s%s", CONFIG_DIR, INI_DENY_FILE);
+	snprintf(file2, sizeof(file2), "/etc/%s", INI_DENY_FILE);
 
-	if ((err = plain_account_init(ACCT_CONFIG_FILE)) < 0)
-		return err;
+	if (!stat(file1, &st) || !stat(file2, &st))
+		log_warning("%s is depreciated and will be obsoleted in the next release, please see README.initiators for more information", INI_DENY_FILE);
 
-	/* TODO: error handling */
+	snprintf(file1, sizeof(file1), "%s%s", CONFIG_DIR, CONFIG_FILE);
+	snprintf(file2, sizeof(file2), "/etc/%s", CONFIG_FILE);
 
-	return err;
-}
-
-static int plain_init(char *params, char **isns, int *isns_ac)
-{
-	FILE *config;
-	char buf[BUFSIZE];
-	char *p, *q;
-
-	if (!(config = fopen(params ? : CONFIG_FILE, "r")))
-		return -errno;
-
-	while (fgets(buf, BUFSIZE, config)) {
-		q = buf;
-		p = target_sep_string(&q);
-		if (!p || *p == '#')
-			continue;
-		if (!strcasecmp(p, "iSNSServer")) {
-			*isns = strdup(target_sep_string(&q));
-		} else if (!strcasecmp(p, "iSNSAccessControl")) {
-			char *str = target_sep_string(&q);
-			if (!strcasecmp(str, "Yes"))
-				*isns_ac = 1;
+	if (!(fp = fopen(params ? params : file1, "r"))) {
+		if ((fp = fopen(file2, "r")))
+			log_warning("%s's location is depreciated and will be moved in the next release to %s", file2, file1);
+		else {
+			log_warning("%s not found, configure through ietadm", CONFIG_FILE);
+			return;
 		}
 	}
 
-	fclose(config);
-	return 0;
+	plain_portal_init(fp, isns, isns_ac);
+
+	rewind(fp);
+
+	plain_target_init(fp);
+
+	rewind(fp);
+
+	plain_account_init(fp);
+
+	fclose(fp);
+
+	return;
 }
 
 struct config_operations plain_ops = {
 	.init			= plain_init,
-	.default_load		= plain_default_load,
 	.target_add		= plain_target_create,
 	.target_del		= plain_target_destroy,
 	.lunit_add		= plain_lunit_create,
@@ -642,5 +719,6 @@ struct config_operations plain_ops = {
 	.account_del		= plain_account_del,
 	.account_query		= plain_account_query,
 	.account_list		= plain_account_list,
-	.initiator_access	= plain_initiator_access,
+	.initiator_allow	= plain_initiator_allow,
+	.target_allow		= plain_target_allow,
 };
